# -*- coding: utf-8 -*-
from typing import List
import datetime as dt

import alchy
import sqlalchemy as sqa

from trailblazer.mip.config import ConfigHandler
from trailblazer.constants import TEMP_STATUSES
from . import models


class BaseHandler:

    User = models.User
    Analysis = models.Analysis
    Job = models.Job
    Info = models.Info

    def setup(self):
        self.create_all()
        # add inital metadata record (for web interface)
        new_info = self.Info()
        self.add_commit(new_info)

    def find_analysis(self, family, started_at, status):
        """Find a single analysis."""
        query = self.Analysis.query.filter_by(
            family=family,
            started_at=started_at,
            status=status,
        )
        return query.first()

    def analyses(self, *, family: str=None, query: str=None, status: str=None, deleted: bool=None,
                 temp: bool=False, before: dt.datetime=None, is_visible: bool=None):
        """Fetch analyses form the database."""
        analysis_query = self.Analysis.query
        if family:
            analysis_query = analysis_query.filter_by(family=family)
        elif query:
            analysis_query = analysis_query.filter(sqa.or_(
                self.Analysis.family.like(f"%{query}%"),
                self.Analysis.status.like(f"%{query}%"),
            ))
        if status:
            analysis_query = analysis_query.filter_by(status=status)
        if isinstance(deleted, bool):
            analysis_query = analysis_query.filter_by(is_deleted=deleted)
        if temp:
            analysis_query = analysis_query.filter(self.Analysis.status.in_(TEMP_STATUSES))
        if before:
            analysis_query = analysis_query.filter(self.Analysis.started_at < before)
        if is_visible is not None:
            analysis_query = analysis_query.filter_by(is_visible=is_visible)
        return analysis_query.order_by(self.Analysis.started_at.desc())

    def analysis(self, analysis_id: int) -> models.Analysis:
        """Get a single analysis."""
        return self.Analysis.query.get(analysis_id)

    def track_update(self):
        """Update the lastest updated date in the database."""
        metadata = self.info()
        metadata.updated_at = dt.datetime.now()
        self.commit()

    def is_running(self, family: str) -> bool:
        """Check if an analysis is currently running/pending for a family."""
        latest_analysis = self.analyses(family=family).first()
        return latest_analysis and latest_analysis.status in TEMP_STATUSES

    def info(self) -> models.Info:
        """Return metadata entry."""
        return self.Info.query.first()

    def add_pending(self, family: str, email: str=None) -> models.Analysis:
        """Add pending entry for an analysis."""
        started_at = dt.datetime.now()
        new_log = self.Analysis(family=family, status='pending', started_at=started_at)
        new_log.user = self.user(email) if email else None
        self.add_commit(new_log)
        return new_log

    def add_user(self, name: str, email: str) -> models.User:
        """Add a new user to the database."""
        new_user = self.User(name=name, email=email)
        self.add_commit(new_user)
        return new_user

    def user(self, email: str) -> models.User:
        """Fetch a user from the database."""
        return self.User.query.filter_by(email=email).first()

    def aggregate_failed(self) -> List:
        """Count the number of failed jobs per category (name)."""
        categories = self.session.query(
            self.Job.name.label('name'),
            sqa.func.count(self.Job.id).label('count')
        ).filter(self.Job.status != 'cancelled').group_by(self.Job.name).all()
        data = [{'name': category.name, 'count': category.count} for category in categories]
        return data

    def jobs(self):
        """Return all jobs in the database."""
        return self.Job.query


class Store(alchy.Manager, BaseHandler, ConfigHandler):

    def __init__(self, uri: str, families_dir: str):
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=models.Model)
        self.families_dir = families_dir



# -*- coding: utf-8 -*-
from flask import abort, g, Blueprint, jsonify, make_response, request
from google.auth import jwt

from trailblazer.server.ext import store

blueprint = Blueprint('api', __name__, url_prefix='/api/v1')


@blueprint.before_request
def before_request():
    if request.method == 'OPTIONS':
        return make_response(jsonify(ok=True), 204)
    auth_header = request.headers.get('Authorization')
    if auth_header:
        jwt_token = auth_header.split('Bearer ')[-1]
    else:
        return abort(403, 'no JWT token found on request')
    user_data = jwt.decode(jwt_token, verify=False)
    user_obj = store.user(user_data['email'])
    if user_obj is None:
        return abort(403, f"{user_data['email']} doesn't have access")
    g.current_user = user_obj


@blueprint.route('/analyses')
def analyses():
    """Display analyses."""
    per_page = int(request.args.get('per_page', 50))
    page = int(request.args.get('page', 1))
    query = store.analyses(status=request.args.get('status'),
                           query=request.args.get('query'),
                           is_visible=request.args.get('is_visible') == 'true' or None)

    query_page = query.paginate(page, per_page=per_page)
    data = []
    for analysis_obj in query_page.items:
        analysis_data = analysis_obj.to_dict()
        analysis_data['user'] = analysis_obj.user.to_dict() if analysis_obj.user else None
        analysis_data['failed_jobs'] = [job_obj.to_dict() for job_obj in analysis_obj.failed_jobs]
        data.append(analysis_data)

    return jsonify(analyses=data)


@blueprint.route('/analyses/<int:analysis_id>', methods=['GET', 'PUT'])
def analysis(analysis_id):
    """Display a single analysis."""
    analysis_obj = store.analysis(analysis_id)
    if analysis_obj is None:
        return abort(404)

    if request.method == 'PUT':
        analysis_obj.update(request.json)
        store.commit()

    data = analysis_obj.to_dict()
    data['failed_jobs'] = [job_obj.to_dict() for job_obj in analysis_obj.failed_jobs]
    data['user'] = analysis_obj.user.to_dict() if analysis_obj.user else None
    return jsonify(**data)


@blueprint.route('/info')
def info():
    """Display meta data about database."""
    metadata_obj = store.info()
    return jsonify(**metadata_obj.to_dict())


@blueprint.route('/me')
def me():
    """Return information about a logged in user."""
    return jsonify(**g.current_user.to_dict())


@blueprint.route('/aggregate/jobs')
def aggregate_jobs():
    """Return stats about jobs."""
    data = store.aggregate_failed()
    return jsonify(jobs=data)



# -*- coding: utf-8 -*-
import datetime as dt
from pathlib import Path
import logging
from typing import List

log = logging.getLogger(__name__)


class FastqHandler:

    @staticmethod
    def name_file(lane: int, flowcell: str, sample: str, read: int,
                  undetermined: bool=False, date: dt.datetime=None, index: str=None) -> str:
        """Name a FASTQ file following MIP conventions."""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date.strftime('%y%m%d') if date else '171015'
        index = index if index else 'XXXXXX'
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_{read}.fastq.gz"

    def link(self, family: str, sample: str, analysis_type: str, files: List[str]):
        """Link FASTQ files for a sample."""
        root_dir = Path(self.families_dir) / family / analysis_type / sample / 'fastq'
        root_dir.mkdir(parents=True, exist_ok=True)
        for fastq_data in files:
            fastq_path = Path(fastq_data['path'])
            fastq_name = self.name_file(
                lane=fastq_data['lane'],
                flowcell=fastq_data['flowcell'],
                sample=sample,
                read=fastq_data['read'],
                undetermined=fastq_data['undetermined'],
            )
            dest_path = root_dir / fastq_name
            if not dest_path.exists():
                log.info(f"linking: {fastq_path} -> {dest_path}")
                dest_path.symlink_to(fastq_path)
            else:
                log.debug(f"destination path already exists: {dest_path}")



import click
from dateutil.parser import parse as parse_date

from trailblazer.store.models import STATUS_OPTIONS


@click.command('ls')
@click.option('-s', '--status', type=click.Choice(STATUS_OPTIONS))
@click.option('-b', '--before', help='return analyses started before date')
@click.pass_context
def ls_cmd(context, before, status):
    """Display recent logs for analyses."""
    runs = context.obj['store'].analyses(
        status=status,
        deleted=False,
        before=parse_date(before) if before else None,
    ).limit(30)
    for run_obj in runs:
        if run_obj.status == 'pending':
            message = f"{run_obj.id} | {run_obj.family} [{run_obj.status.upper()}]"
        else:
            message = (f"{run_obj.id} | {run_obj.family} {run_obj.started_at.date()} "
                       f"[{run_obj.type.upper()}/{run_obj.status.upper()}]")
            if run_obj.status == 'running':
                message = click.style(f"{message} - {run_obj.progress * 100}/100", fg='blue')
            elif run_obj.status == 'completed':
                message = click.style(f"{message} - {run_obj.completed_at}", fg='green')
            elif run_obj.status == 'failed':
                message = click.style(message, fg='red')
        print(message)



# -*- coding: utf-8 -*-
from copy import deepcopy
import logging
from pathlib import Path

from marshmallow import Schema, fields, validate
import ruamel.yaml

from trailblazer.constants import DEFAULT_CAPTURE_KIT
from trailblazer.exc import ConfigError

LOG = logging.getLogger(__name__)


class SampleSchema(Schema):
    sample_id = fields.Str(required=True)
    analysis_type = fields.Str(
        required=True,
        validate=validate.OneOf(choices=['wes', 'wgs', 'tga'])
    )
    father = fields.Str(default='0')
    mother = fields.Str(default='0')
    phenotype = fields.Str(
        required=True,
        validate=validate.OneOf(choices=['affected', 'unaffected', 'unknown'])
    )
    sex = fields.Str(
        required=True,
        validate=validate.OneOf(choices=['male', 'female', 'unknown'])
    )
    expected_coverage = fields.Float()
    capture_kit = fields.Str(default=DEFAULT_CAPTURE_KIT)


class ConfigSchema(Schema):
    family = fields.Str(required=True)
    default_gene_panels = fields.List(fields.Str(), required=True)
    samples = fields.List(fields.Nested(SampleSchema), required=True)


class ConfigHandler:

    def make_config(self, data: dict):
        """Make a MIP config."""
        self.validate_config(data)
        config_data = self.prepare_config(data)
        return config_data

    @staticmethod
    def validate_config(data: dict) -> dict:
        """Convert to MIP config format."""
        errors = ConfigSchema().validate(data)
        if errors:
            for field, messages in errors.items():
                if isinstance(messages, dict):
                    for level, sample_errors in messages.items():
                        sample_id = data['samples'][level]['sample_id']
                        for sub_field, sub_messages in sample_errors.items():
                            LOG.error(f"{sample_id} -> {sub_field}: {', '.join(sub_messages)}")
                else:
                    LOG.error(f"{field}: {', '.join(messages)}")
            raise ConfigError('invalid config input', errors=errors)

    @staticmethod
    def prepare_config(data: dict) -> dict:
        """Prepare the config data."""
        data_copy = deepcopy(data)
        # handle single sample cases with 'unknown' phenotype
        if len(data_copy['samples']) == 1:
            if data_copy['samples'][0]['phenotype'] == 'unknown':
                LOG.info("setting 'unknown' phenotype to 'unaffected'")
                data_copy['samples'][0]['phenotype'] = 'unaffected'
        # set the mother/father to '0' if they are not set for a sample
        for sample_data in data_copy['samples']:
            sample_data['mother'] = sample_data.get('mother') or '0'
            sample_data['father'] = sample_data.get('father') or '0'
            if sample_data['analysis_type'] == 'wgs' and sample_data.get('capture_kit') is None:
                sample_data['capture_kit'] = DEFAULT_CAPTURE_KIT
        return data_copy

    def save_config(self, data: dict) -> Path:
        """Save a config to the expected location."""
        out_dir = Path(self.families_dir) / data['family']
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = out_dir / 'pedigree.yaml'
        dump = ruamel.yaml.round_trip_dump(data, indent=4, block_seq_indent=2)
        out_path.write_text(dump)
        return out_path



# -*- coding: utf-8 -*-
"""Test MIP files"""

from trailblazer.mip import trending


def test_parse_mip_analysis(files_raw) -> dict:
    """
    Args:
    files_raw (dict): With dicts from files
    """
    # GIVEN config data of a "sharp" run (not dry run)
    mip_config_raw = files_raw['config']

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw['sampleinfo']

    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw['qcmetrics']

    trend_data = trending.parse_mip_analysis(mip_config_raw, qcmetrics_raw, sampleinfo_raw)

    ## Build dict for family return data
    trend_test_data = {
        'family': 'family',
        'genome_build': 'GRCh37',
        'mip_version': 'v6.0.0',
        'rank_model_version': '1.20'
    }

    # Check returns from def
    for key, value in trend_test_data.items():
        assert trend_data[key] == value

    ## Check sample id
    assert 'sample' in trend_data['sample_ids']

    ## Build dict for sample return data
    trend_test_sample_data = {
        'at_dropout': {
            'sample': 0.126963,
            },
        'analysis_sex': {
            'sample': 'female',
            },
        'duplicates': {
            'sample': 13.2719986683768,
            },
        'insert_size_standard_deviation': {
            'sample': 89.871783,
            },
        'mapped_reads': {
            'sample': 98.50732625744484,
            },
        'median_insert_size': {
            'sample': 413,
            },
        'gc_dropout': {
            'sample': 3.956909,
            },
    }

    # Check returns from def
    for metric in trend_test_sample_data:
        for sample, value in trend_test_sample_data[metric].items():
            assert trend_data[metric][sample] == value
